// Search.java
// javac Search.java

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Locale;
import java.util.Vector;
import java.util.StringTokenizer;
import javax.help.*;
import javax.help.event.*;
import javax.help.search.*;

public class Search extends Applet 
implements ActionListener, SearchListener, ItemListener
{
   private Button        searchButton       = null;
   private TextField     criteria           = new TextField();
   private TextField     message            = new TextField();
   private List          list               = new List();
   private Vector        finalVector        = null;
   private Vector        searchDatabases    = new Vector();
   private Vector        searchItemVector   = null;
   private Vector        searchQueries      = new Vector();
   private boolean       startSearch        = false;
   private String        buttonText         = null;
   private String        searchStarted      = null;
   private String        searchEnded        = null;
   private String        nothingFound       = null;
   private String        name               = null;
   private String        bkColor            = null;
   private int           goBack             = 0;
   private int           searchQueriesNum   = 0;
   private int           queryEventCounter  = 0;

   public void init()
   {
      // Get national language dependent strings.
      buttonText = getParameter("buttontext"); // text on button
      if (buttonText == null)
         buttonText = "search";
      searchButton = new Button(buttonText);
      searchStarted = getParameter("searchstarted"); // message when search started
      if (searchStarted == null)
         searchStarted = "search started";
      searchEnded = getParameter("searchended"); // message when search ended, something found
      if (searchEnded == null)
         searchEnded = "search ended";
      nothingFound = getParameter("nothingfound"); // message when search ended , nothing found
      if (nothingFound == null)
         nothingFound = "nothing found";

      // Get other environmental information from invoker.
      name = getParameter("name"); // name attribute of frame to display HTML
      if (name == null)
         name = "main";
      bkColor = getParameter("bkColor"); // hex background color of search applet (eg, "FFFFFF")
      String searchDBDir = getParameter("searchdbdir"); // search DB directory, relative to applet
      if (searchDBDir == null)
         searchDatabases.addElement("JavaHelpSearch"); // default produced by JavaHelp Indexer
      else
      {
         // Look for several search databases, in comma delimited string
         StringTokenizer st = new StringTokenizer(searchDBDir, ",", false);
         while (st.hasMoreTokens())
         {
            searchDatabases.addElement((st.nextToken()).trim());
         }
      }
      searchDatabases.trimToSize();

      // Find out how many directories to backup, if necessary, to get from
      // applet codebase to document root directory, if applet not in doc root directory.
      // If there are several search databases, then backup will automatically be zero.
      String testGoBack = getParameter("goback");
      int testInt = 0;
      if (searchDatabases.size() == 1)
      {
         try
         {
            testInt = Integer.parseInt(testGoBack);
         }
         catch (Exception e)
         {
            testInt = 0;
         }
      }
      goBack = Math.max(0, testInt);

      // Set background color for applet
      if (bkColor != null)
      {
         try
         {
            this.setBackground(new Color(Integer.parseInt(bkColor, 16)));
         }
         catch (Exception e)
         {/*nothing - default color will be used*/}
      }

      // Start arranging components
      this.setLayout(new BorderLayout());

      Panel p1 = new Panel(new BorderLayout());
      p1.add(searchButton, "West");
      p1.add(criteria, "Center");

      Panel p2 = new Panel(new GridLayout(1, 1));
      p2.add(message);
      message.setEditable(false);

      Panel p3 = new Panel(new BorderLayout());
      p3.add(p1, "North");
      p3.add(p2, "South");

      this.add(p3, "North");
      this.add(list,"Center");

      // Set up parms to get Sun's default search engine
      Hashtable params = new Hashtable();
      SearchEngine searchEngine = null;
      URL base = null;
      try
      {
         base = getCodeBase();
      }
      catch (Exception e)
      {
         System.out.println("Problem getting codebase: " + e);
      }
      params.put("engine", "com.sun.java.help.search.DefaultSearchEngine");

      // Create a SearchQuery object for each search database
      for (int i = 0; i < searchDatabases.size(); i++)
      {
         params.put("data", (String)searchDatabases.elementAt(i));
         try
         {
            // Try to create a new search query based on the search engine & search databases,
            // but ignore any problems that arise (say, search DB does not exist).  Continue
            // witb valid search databases, which result in an entry into searchQueries Vector.
            searchEngine = new com.sun.java.help.search.DefaultSearchEngine(base, params);
            SearchQuery searchQuery = searchEngine.createQuery();
            searchQueries.addElement(searchQuery);
            searchQuery.addSearchListener(this);  // add listener for each query
         }
         catch (Exception e)
         {
            System.out.println("Problem creating search engine/query, database: " + 
                               (String)searchDatabases.elementAt(i) + 
                               ", Exception: " + e);
         }
      }
      searchQueries.trimToSize();
      searchQueriesNum = searchQueries.size();

      // add remaining listeners...
      searchButton.addActionListener(this);
      list.addItemListener(this);
      criteria.addActionListener(this);
   }

   public void start()
   {
      criteria.requestFocus(); // place cursor where user enters search criteria
   }

   // --- implement SearchListener methods ---
   public synchronized void itemsFound(SearchEvent e)
   {
      // If we are starting a series of searches, begin with a new Vector of search events.
      if (startSearch == true)
      {
         searchItemVector = new Vector();
         startSearch = false;
         queryEventCounter = 0;
      }

      // Accumulate the search items for each query event
      Enumeration itemEnum = e.getSearchItems();
      while (itemEnum.hasMoreElements())
      {
         searchItemVector.addElement(itemEnum.nextElement());
      }
      queryEventCounter++;

      if (queryEventCounter == searchQueriesNum)
         processSearchItems(searchItemVector.elements());
   }

   public void searchStarted(SearchEvent e)
   {
      // set message for user
      message.setText(searchStarted);
   }

   public void searchFinished(SearchEvent e)
   {
      // set message for user, only if last query has been processed.
      if (queryEventCounter == searchQueriesNum)
      {
         if (finalVector.isEmpty())
            message.setText(nothingFound);
         else
            message.setText(searchEnded);
      }
   }

   // --- implement ActionListner interface
   public void actionPerformed(ActionEvent e)
   {
      if ((e.getSource() == searchButton) || (e.getSource() == criteria))
      {
         // User pressed search button - start a search, or multiple
         // searches if there are multiple search databases.  
         synchronized(this)
         {
            startSearch = true;
            for (int i = 0; i < searchQueriesNum; i++)
            {
               ((SearchQuery)(searchQueries.elementAt(i))).start(criteria.getText(), this.getLocale());
            }
         }
      }
   }
   // --- implement ItemListener interface ---
   // user has selected in hit listbox - display corresponding HTML
   public void itemStateChanged(ItemEvent e)
   {
      if (e.getStateChange() == ItemEvent.SELECTED)
      {
         int index = list.getSelectedIndex(); 
         try
         {
            // Create a URL, possibly backing up from applet codebase to the
            // document root directory, and tacking on the HTML file name
            URL url = new URL(backupUrl(getCodeBase(), goBack) + 
                              ((LineItem)finalVector.elementAt(index)).getFilename());
            getAppletContext().showDocument(url, name);
         }
         catch (Exception e1)
         {
            System.out.println("Error attempting to display HTML: " + e1);
         }
      }
   }

   // ---  helper method, varying number of *'s depending on confidence ---
   private String getHML(double confidence)
   {
      if (confidence < 1)
         return "*****";
      else if (confidence < 5)
         return "**** ";
      else if (confidence < 12.5)
         return "***  ";
      else if (confidence < 25)
         return "**   ";
      else
         return "*    ";
   }

   // --- helper method, back up specified number of directories ---
   private URL backupUrl(URL url, int num) throws MalformedURLException
   {
      String stringUrl = url.toString();

      for (int i = 0; i < num; i++)
      {
         int fromIndex = Math.max(0, stringUrl.length()-2); // look from left of rightmost '/'
         int newIndex = stringUrl.lastIndexOf('/', fromIndex); // find second '/' from right
         // Now chop off a directory, until required number are gone.
         if (newIndex > 0)
         {
            stringUrl = stringUrl.substring(0, newIndex+1);  
         }
      }
      return new URL(stringUrl);
   }

   // --- helper method, process combined search items from (possibly) multiple queries
   private void processSearchItems(Enumeration itemEnum)
   {
      // Next two lines seem needed to "prime" previous hit list so that
      // next hit list will display properly.
      if (list.getItemCount() > 0)
         list.select(0);

      // May be mulitple hits or search items on the same HTML file. - this
      // is what the search engine returns.  Save hits for processing.  The
      // qualified file name of the HTML file (relative to the base variable)
      // makes it unique and serves to identify multiple hits on the same
      // HTML file.
      Hashtable itemHash = new Hashtable();
      while (itemEnum.hasMoreElements())
      {
         SearchItem item = (SearchItem)itemEnum.nextElement();
         if (itemHash.get(item.getFilename()) != null)
         {
            // An existing hit on the same HTML file...
            LineItem temp = (LineItem)itemHash.get(item.getFilename());
            int currentHits = temp.getHits();
            double currentConfidence = temp.getConfidence();
            double newConfidence = Math.min(currentConfidence, item.getConfidence());
            temp.setHits(++currentHits);
            temp.setConfidence(newConfidence);
         }
         else
         {
            itemHash.put(item.getFilename(), 
                         new LineItem(item.getBase(), item.getFilename(),
                                      item.getTitle(), item.getConfidence()));
         }
      }

      finalVector = new Vector();
      Enumeration test = itemHash.elements();
      while (test.hasMoreElements())
      {
         LineItem temp = (LineItem)test.nextElement();
         if (finalVector.isEmpty())
         {
            // put first element into vector
            finalVector.addElement(temp);
         }
         else
         {
            // compare successive elements in hashtable to all
            // elements currently in finalVector.
            int size = finalVector.size();
            boolean keepGoing = true;
            for (int i = 0; i < size && keepGoing; i++)
            {
               // Given an item from the hashtable... 

               // If confidence is less ("better") then insert, shifting other
               // items down, otherwise, do nothing.

               // If confidence same, test on number of hits.  If hits more, 
               // then insert, shifting others down, otherwise, do nothing.

               // Otherwise, just put at end if not otherwise added.

               if (temp.getConfidence() < 
                   ((LineItem)finalVector.elementAt(i)).getConfidence())
               {
                  finalVector.insertElementAt(temp, i);
                  keepGoing = false;
               }
               else if (temp.getConfidence() == 
                        ((LineItem)finalVector.elementAt(i)).getConfidence() &&
                        temp.getHits() > ((LineItem)finalVector.elementAt(i)).getHits())
               {
                  finalVector.insertElementAt(temp, i);
                  keepGoing = false;
               }
               else if (i == (size-1))
               {
                  finalVector.addElement(temp);
                  keepGoing = false;
               }
            }
         }
      }

      // Now populate selection list...
      list.removeAll();
      Enumeration listEnum = finalVector.elements();
      while (listEnum.hasMoreElements())
      {
         LineItem item = (LineItem)listEnum.nextElement();
         list.add("(" + getHML(item.getConfidence()) + item.getHits() + ") " + item.getTitle());
      }
      // Trigger HTML display of first item in selection list
      if (list.getItemCount() > 0)
      {
         list.select(0);
         ItemEvent evt = new ItemEvent(list, 
                                       ItemEvent.ITEM_STATE_CHANGED, 
                                       new Integer(0), 
                                       ItemEvent.SELECTED);
         itemStateChanged(evt);
      }
   }
}



